#!/usr/bin/env python

# $Id: tmutils.py,v 1.8 1999/07/12 07:45:16 grove Exp $

# ----------------------------------------------------------------------------
#
# tmproc - a Topic Map processor
#
# Copyright (C) 1999 Geir O. Grnmo <grove@infotek.no>
#
# ----------------------------------------------------------------------------

"""A collection of Topic Map utilities.

Includes classes for import and exporting Topic Maps. A class for printing
statistical information is also included."""

__author__ = "Geir Ove Grnmo, grove@infotek.no"
__version__ = "0.20"

# ----------------------------------------------------------------------------
# TODO:
#
# General:
#
# - Add validation of topic map document and constraints
# - Make sure that the #GI mapping defaulting on the various tm attributes works
# - Finish the TMExporter, which exports tm objects to the interchange
#   format. This should be tested if it return the same result on import
#   after an export.
# - add support for entity declarations in the importer and exporter
# - HyTime stuff is ignored
# - Support the addthms element form
# - Use listeners instead of one single topic map module!
# - Support for merging Topic Maps
# - Add correct support for locators [queryloc]

# ============================================================================
# Imported modules
# ============================================================================

from xml.sax import saxexts, saxlib, saxutils
import string, sys

# ============================================================================
# Class that implements helper methods for parsing Topic Maps
# ============================================================================

class TMUtils:

    def __init__(self, tm_handler):

        """Initialize a new instance."""

	self.tm = None
	
	self.tm_handler = tm_handler
	self.arch_handler = None
	self.arch_name = None

    # ----------------------------------------------------------------------------
    # Document handlers
    # ----------------------------------------------------------------------------

    def set_tm_handler(self, tm_handler):

	"""Set the Topic Map document handler to use."""

	self.tm_handler = tm_handler

    def get_tm_handler(self):

	"""Get the current Topic Map document handler."""
	
	return self.tm_handler

    def set_arch_handler(self, arch_name, arch_handler):
	
	"""Set the architectural forms document handler to use."""

	self.arch_name = arch_name
	self.arch_handler = arch_handler
	
    def get_arch_handler(self):

	"""Get the current architectural forms document handler."""

	return self.arch_handler
    
    # ----------------------------------------------------------------------------
    # Convenient document parsing methods
    # ----------------------------------------------------------------------------

    def get_parser(self, validating=0):

	"Get a parser object by using the standard saxexts ParserFactory."

	if validating:
	    pf=saxexts.XMLValParserFactory
	else:
	    pf=saxexts.XMLParserFactory

	return pf.make_parser()

    def get_parser_by_driver(self, driver):

	"Get a parser object by using the name of the parser's SAX driver."

	return saxexts.ParserFactory([driver]).make_parser()

    def __init_parser(self, parser):

	"""Initialize the parser and the document handlers."""

	if self.arch_handler:
	    # Register tm document handler with architecture processor
	    self.arch_handler.add_document_handler(self.arch_name, self.tm_handler)
	    self.tm_handler.set_arch_handler(self.arch_handler)
	    # Register architecture document handler
	    parser.setDocumentHandler(self.arch_handler)
	else:
	    # Register tm document document handler
	    parser.setDocumentHandler(self.tm_handler)

	parser.setErrorHandler(saxutils.ErrorPrinter())
	
    def parse(self, parser, system_id):

	"""Build a Topic Map datastructure from the document specified in the system identifier."""
	
	self.__init_parser(parser)
	parser.parse(system_id)
	self.tm = self.tm_handler.get_topic_map()
	return self.tm

    def parseFile(self, parser, file_obj):

	"""Build a Topic Map datastructure from the file object."""

	self.__init_parser(parser)
	parser.parseFile(file_obj)

	self.tm = self.tm_handler.get_topic_map()
	return self.tm

# ============================================================================
# Topic Map import
# ============================================================================

class TMImporter(saxlib.HandlerBase):

    """A SAX document handler that produces a Topic Map datastructure from a SAX event stream.

    The input data must conform to the ISO/IEC 13250:1999 interchange
    format.

    An instance of this class should be passed as a SAX
    DocumentHandler to a parser driver that implements the SAX
    interfaces.  The Topic Map modules only have to implement the
    one-object-set methods and the add methods.

    Implicit topics are created automatically. They only contain their
    sgml id by default ."""

    def __init__(self, topic_map_name, tm_module_name, tm_module_keywords):

        "Initialize a new instance."

	self.tm_module = self._get_module(tm_module_name)
	self.factory = self.tm_module.TopicMapFactory(tm_module_keywords)

	self.topic_map_name = topic_map_name

	self.arch_handler = None
	
	self.tm = None
	self.topic = None
	self.topic_name = None
	self.occurrence = None
	self.association = None
	self.assoc_role = None
	self.facet = None
	self.facet_value = None

	self.elem_stack = []
	
	self.topic_basename = None
	self.topic_dispname = None
	self.topic_sortname = None
	self.topic_nametext = ""
	
	self.occur_addresses = ""
	self.assoc_role_addresses = ""
	self.facet_value_addresses = ""

	self.occur_loctype = ""
	self.assoc_role_loctype = ""
	self.facet_value_loctype = ""

    def set_tm_module_by_name(self, tm_module_name):

	self.tm_module = self._get_module(tm_module_name)

    def set_tm_module(self, tm_module):

	self.tm_module = tm_module

    def _get_module(self, module_name):

	if sys.platform[:4] == "java":
	    from org.python.core import imp
	    return imp.importName(module_name, 0, globals())
	else:
	    import imp
	    fp, pathname, description = imp.find_module(module_name)
	    try:
		return imp.load_module(module_name, fp, pathname, description)
	    finally:
		if fp:
		    fp.close()
		    
    def get_topic_map(self):

	"""Get the imported TopicMap object."""
	
	return self.tm

    def set_arch_handler(self, arch_handler):

	"""Set the architecture document handler that is used for architectural processing."""
	
	self.arch_handler = arch_handler
	
    # ----------------------------------------------------------------------------
    # SAX events

    def processingInstruction (self, target, remainder):

	"Handle a processing instruction event."
	
    def startElement(self, name, amap):

	"Handle an event for the beginning of an element."
	
	# Store parse context [elements]
	self.elem_stack.append(name)
	
	# -----------------------------------------------------------------------------
	# Topic Map
	# -----------------------------------------------------------------------------
	if name == "topicmap":
	    self.tm = self.factory.make_topic_map(self.topic_map_name)
	    for attr in amap.keys():
		if attr == "addthems":
		    # No default
		    for theme in string.split(amap[attr]):
			self.tm.add_added_theme(self.factory.make_topic(self.tm, theme))

	# -----------------------------------------------------------------------------
	# Topic
	# -----------------------------------------------------------------------------
	elif name == "topic":
	    sgml_id=""
	    if amap.has_key("id"): sgml_id = amap["id"]
	    self.topic = self.factory.make_topic(self.tm, sgml_id)

	    for attr in amap.keys():
		if attr == "id":
		    self.topic.set_sgml_id(sgml_id)
		elif attr == "identity":
		    self.topic.set_identity(amap[attr])
		elif attr == "linktype":
		    # Default: GI
		    self.topic.set_linktype(amap[attr])
		elif attr == "types":
		    for type in string.split(amap[attr]):
			self.topic.add_type(self.factory.make_topic(self.tm, type))
		elif attr == "scope":
		    for scope in string.split(amap[attr]):
			self.topic.add_scope(self.factory.make_topic(self.tm, scope))

	    # Set the default link type
	    if not self.topic.get_linktype() and self.arch_handler:
		self.topic.set_linktype(self.arch_handler.get_current_element_name())
		
	# -----------------------------------------------------------------------------
	# Topic Name
	# -----------------------------------------------------------------------------
	elif name == "topname":
	    # FIXME:	    if not self.topic: return
	    self.topic_name = self.factory.make_topic_name(self.tm)
	    for attr in amap.keys():
		if attr == "scope":
		    # Default: scope is unconstrained
		    for scope in string.split(amap[attr]):
			self.topic_name.add_scope(self.factory.make_topic(self.tm, scope))

	elif name == "basename":
	    self.topic_basename = self.factory.make_name(self.tm)
	    for attr in amap.keys():
		if attr == "scope":
		    # Default: scope is unconstrained
		    for scope in string.split(amap[attr]):
			self.topic_basename.add_scope(self.factory.make_topic(self.tm, scope))
	    
	elif name == "dispname":	
	    self.topic_dispname = self.factory.make_name(self.tm)
	    for attr in amap.keys():
		if attr == "scope":
		    # Default: scope is unconstrained
		    for scope in string.split(amap[attr]):
			self.topic_dispname.add_scope(self.factory.make_topic(self.tm, scope))

	elif name == "sortname":	
	    self.topic_sortname = self.factory.make_name(self.tm)
	    for attr in amap.keys():
		if attr == "scope":
		    # Default: scope is unconstrained
		    for scope in string.split(amap[attr]):
			self.topic_sortname.add_scope(self.factory.make_topic(self.tm, scope))

	# -----------------------------------------------------------------------------
	# Occurence
	# -----------------------------------------------------------------------------
	elif name == "occurs":
	    self.occurrence = self.factory.make_occurrence(self.tm)
	    for attr in amap.keys():
		if attr == "occrl":
		    # Default: GI
		    self.occurrence.set_rolename(amap[attr])
		elif attr == "type":
		    self.occurrence.set_type(self.factory.make_topic(self.tm, amap[attr]))
		elif attr == "scope":
		    # Default: rolename [of public subject OCCTYPE] or GI
		    for scope in string.split(amap[attr]):
			self.occurrence.add_scope(self.factory.make_topic(self.tm, scope))
		elif attr == "loctype":
		    self.occur_loctype = amap[attr]

	    # Set the default occurrence role
	    if not self.occurrence.get_rolename() and self.arch_handler:
		self.occurrence.set_rolename(self.arch_handler.get_current_element_name())

	# -----------------------------------------------------------------------------
	# Association
	# -----------------------------------------------------------------------------
	elif name == "assoc":
	    self.association = self.factory.make_association(self.tm)
	    for attr in amap.keys():
		if attr == "linktype":
		    self.association.set_linktype(amap[attr])
		elif attr == "scope":
		    # Default: scope is unconstrained
		    for scope in string.split(amap[attr]):
			self.association.add_scope(self.factory.make_topic(self.tm, scope))
		elif attr == "type":
		    self.association.set_type(self.factory.make_topic(self.tm, amap[attr]))
		    
	    # Set the default link type
	    if not self.association.get_linktype() and self.arch_handler:
		self.association.set_linktype(self.arch_handler.get_current_element_name())

	# -----------------------------------------------------------------------------
	# Association role
	# -----------------------------------------------------------------------------
	elif name == "assocrl":
	    self.assoc_role = self.factory.make_association_role(self.tm)
	    for attr in amap.keys():
		if attr == "anchrole":
		    # Default GI
		    self.assoc_role.set_rolename(amap[attr])
		elif attr == "type":
		    self.assoc_role.set_type(self.factory.make_topic(self.tm, amap[attr]))
		elif attr == "loctype":
		    self.assoc_role_loctype = amap[attr]
#EDF added to support XTM files
                elif attr == "href":
  	            self.assoc_role_addresses = amap[attr]

	    # Set the default anchor role
	    if not self.assoc_role.get_rolename() and self.arch_handler:
		self.assoc_role.set_rolename(self.arch_handler.get_current_element_name())

	# -----------------------------------------------------------------------------
	# Facet
	# -----------------------------------------------------------------------------
	elif name == "facet":
	    self.facet = self.factory.make_facet(self.tm)
	    for attr in amap.keys():
		if attr == "linktype":
		    # Default: GI
		    self.facet.set_linktype(amap[attr])
		elif attr == "type":
		    self.facet.set_type(self.factory.make_topic(self.tm, amap[attr]))

	    # Set the default type
	    if not self.facet.get_linktype() and self.arch_handler:
		self.facet.set_linktype(self.arch_handler.get_current_element_name())

	# -----------------------------------------------------------------------------
	# Facet value
	# -----------------------------------------------------------------------------
	elif name == "fvalue":
	    self.facet_value = self.factory.make_facet_value(self.tm)
	    for attr in amap.keys():
		if attr == "facetval":
		    # Default GI
		    self.facet_value.set_value(amap[attr])
		elif attr == "type":
		    self.facet_value.set_type(self.factory.make_topic(self.tm, amap[attr]))
		elif attr == "loctype":
		    self.facet_value_loctype = amap[attr]

	    # Set the default value
	    if not self.facet_value.get_value() and self.arch_handler:
		self.facet.set_value(self.arch_handler.get_current_element_name())

    def endElement(self, name):

	"Handle an event for the end of an element."

	# Topic Map
	if name == "topicmap":
	    self.tm.finalize()
	
	# Topic
	elif name == "topic":
	    self.topic.finalize()

	    self.topic = None
	    
	# Topic Name
	elif name == "topname":
	    self.topic.add_name(self.topic_name)
	    self.topic_name.finalize()

	    self.topic_name = None

	elif name == "basename":
	    self.topic_basename.set_name(self.topic_nametext)
	    self.topic_name.add_basename(self.topic_basename)
	    self.topic_basename = None
	    self.topic_nametext = ""

	elif name == "dispname":	
	    self.topic_dispname.set_name(self.topic_nametext)
	    self.topic_name.add_dispname(self.topic_dispname)
	    self.topic_dispname = None
	    self.topic_nametext = ""

	elif name == "sortname":	
	    self.topic_sortname.set_name(self.topic_nametext)
	    self.topic_name.add_sortname(self.topic_sortname)
	    self.topic_sortname = None
	    self.topic_nametext = ""

	# Occurrence
	elif name == "occurs":
	    for address in string.split(self.occur_addresses):
		self.occurrence.add_locator(self.factory.make_locator(self.tm, address, self.occur_loctype))
	    self.topic.add_occurrence(self.occurrence)
	    self.occurrence.finalize()
	    
	    self.occurrence = None
	    self.occur_addresses = ""
	    self.occur_loctype = None

	# Association
	elif name == "assoc":
	    self.association.finalize()
	    
	    self.association = None

	# Association role
	elif name == "assocrl":
	    for address in string.split(self.assoc_role_addresses):
		#self.assoc_role.add_locator(self.factory.make_locator(self.tm, address, self.assoc_role_loctype))
		self.assoc_role.add_locator(self.factory.make_topic(self.tm, address, self.assoc_role_loctype))

	    self.association.add_role(self.assoc_role)
	    self.assoc_role.finalize()

#EDF added to assign types to topics
	    for address in string.split(self.assoc_role_addresses):
              if self.tm.is_type(self.tm[address], self.assoc_role.get_type()) == 0:
                self.tm[address].add_type(self.assoc_role.get_type())

	    self.assoc_role = None
	    self.assoc_role_addresses = ""
	    self.assoc_role_loctype = None

	# Facet
	elif name == "facet":
	    self.facet.finalize()

	    self.facet = None

	# Facet value
	elif name == "fvalue":
	    for address in string.split(self.facet_value_addresses):
		#self.facet_value.add_locator(self.factory.make_locator(self.tm, address, self.facet_value_loctype))
		self.facet_value.add_locator(self.factory.make_topic(self.tm, address, self.facet_value_loctype))

	    self.facet.add_value(self.facet_value)
	    self.facet.finalize()

	    self.facet_value = None
	    self.facet_value_addresses = ""
	    self.facet_value_loctype = None

	# Pop this element off the context list
	del self.elem_stack[-1]
	
    def characters(self, data, start_ix, length):

	"Handle a character data event."

	# -----------------------------------------------------------------------------
	# Topic names
	# -----------------------------------------------------------------------------

	# FIXME: Beware of splitted characters events
	
	# Topic base name
	if self.elem_stack[-1] == "basename":
	    if not self.topic or not self.topic_name: return
	    self.topic_nametext = self.topic_nametext + string.strip(data[start_ix:start_ix+length])
	# Topic display name
	elif self.elem_stack[-1] == "dispname":
	    if not self.topic or not self.topic_name: return
	    self.topic_nametext = self.topic_nametext + string.strip(data[start_ix:start_ix+length])
	# Topic display name
	elif self.elem_stack[-1] == "sortname":
	    if not self.topic or not self.topic_name: return
	    self.topic_nametext = self.topic_nametext + string.strip(data[start_ix:start_ix+length])
	# Topic occurrence locators
	elif self.elem_stack[-1] == "occurs":
	    self.occur_addresses = self.occur_addresses + data[start_ix:start_ix+length]
	# Association role locators
	elif self.elem_stack[-1] == "assocrl":
	    self.assoc_role_addresses = self.assoc_role_addresses + data[start_ix:start_ix+length]
	# Facet value locators
	elif self.elem_stack[-1] == "fvalue":
	    self.facet_value_addresses = self.facet_value_addresses + data[start_ix:start_ix+length]

	# FIXME: Handle locators for Topic occurrence, Association role and Facet value
	
    def ignorableWhitespace(self, data, start_ix, length):

	"Handle an event for ignorable whitespace in element content."

    def startDocument(self):
	
	"Handle an event for the beginning of a document."

    def endDocument(self):

	"Handle an event for the end of a document."

# ============================================================================
# Topic Map export
# ============================================================================

class TMExporter:

    """Class that exports TM information to the TM interchange format.

    The output is done through the SAX DocumentHandler interface. Attach a
    document handler that outputs the appropriate output. E.g. use the
    xmlarch.Prettifier class to get 'prettified' output."""
    
    def __init__(self):

	pass

    def __get_names_string(self, topics):

	ids = []
	for topic in topics:
	    ids.append(topic.get_sgml_id())
	return string.join(ids)
    
    def export(self, topic_map, document_handler):

	tm = topic_map
	dh = document_handler
	
	dh.startDocument()

	attrs={}
	if tm.get_added_themes(): attrs["addthems"] = self.__get_names_string(tm.get_added_themes())
	dh.startElement("topicmap", attrs)

	# -----------------------------------------------------------------------------
	# Topics
	# -----------------------------------------------------------------------------

	for topic in tm.get_topics():
	    attrs = {}
	    if topic.get_sgml_id(): 
                attrs["id"] = topic.get_sgml_id()
	    if topic.get_identity():
		attrs["identity"] = topic.get_identity()
#EDF fixed attribute name
	    if topic.get_types(): 
                attrs["types"] = self.__get_names_string(topic.get_types())
		    
	    dh.startElement("topic", attrs)

	    # Topic names
	    for name in topic.get_names():
		attrs = {}
		if name.get_scopes(): 
                   attrs["scope"] = self.__get_names_string(name.get_scopes())
		    
		dh.startElement("topname", attrs)

		# Basename
		for basename in name.get_basenames():
		    _name = basename.get_name()
		    dh.startElement("basename", {})
		    dh.characters(_name, 0, len(_name))
		    dh.endElement("basename")

		# Display names
		for dispname in name.get_dispnames():
		    _name = dispname.get_name()
		    dh.startElement("dispname", {})
		    dh.characters(_name, 0, len(_name))
		    dh.endElement("dispname")

		# Sort names
		for sortname in name.get_sortnames():
		    _name = sortname.get_name()
		    dh.startElement("sortname", {})
		    dh.characters(_name, 0, len(_name))
		    dh.endElement("sortname")

		dh.endElement("topname")

	    # Occurrences
	    for occur in topic.get_occurrences():
		attrs = {}
		if occur.get_rolename(): attrs["occrl"] = occur.get_rolename()
		if occur.get_type():attrs["type"] = occur.get_type().get_sgml_id()
		if occur.get_scopes():attrs["scope"] = self.__get_names_string(occur.get_scopes())
		    
		dh.startElement("occurs", attrs)
		locators = []
		for locator in occur.get_locators():
		    locators.append(locator.get_address())
		locator_string = string.join(locators)
		dh.characters(locator_string, 0, len(locator_string))
		dh.endElement("occurs")

	    dh.endElement("topic")

	# -----------------------------------------------------------------------------
	# Associations
	# -----------------------------------------------------------------------------

	for assoc in tm.get_associations():
	    attrs = {}
	    if assoc.get_scopes(): attrs["scope"] = self.__get_names_string(assoc.get_scopes())
	    if assoc.get_type(): attrs["type"] = assoc.get_type().get_sgml_id()
		
	    dh.startElement("assoc", attrs)

	    # Association roles
	    for assoc_role in assoc.get_roles():
		attrs = {}
		if assoc_role.get_rolename(): attrs["anchrole"] = assoc_role.get_rolename()
		if assoc_role.get_type(): attrs["type"] = assoc_role.get_type().get_sgml_id()
		dh.startElement("assocrl", attrs)
		
		locators = []
		for locator in assoc_role.get_locators():
		    # FIXME: locator must in this case be a topic
		    locators.append(locator.get_sgml_id())
		locator_string = string.join(locators)
		dh.characters(locator_string, 0, len(locator_string))
		dh.endElement("assocrl")

	    dh.endElement("assoc")

	# -----------------------------------------------------------------------------
	# Facets
	# -----------------------------------------------------------------------------

	for facet in tm.get_facets():
	    attrs={}
	    if facet.get_type(): attrs["type"] = facet.get_type().get_sgml_id()
	    dh.startElement("facet", attrs)

	    # Facet values
	    for fvalue in facet.get_values():
		attrs = {}
		if fvalue.get_value(): attrs["facetval"] = fvalue.get_value()
		if fvalue.get_type(): attrs["type"] = fvalue.get_type().get_sgml_id()
		dh.startElement("fvalue", attrs)
		locators = []
		for locator in fvalue.get_locators():
		    # FIXME: locator must in this case be a topic
		    locators.append(locator.get_sgml_id())
		locator_string = string.join(locators)
		dh.characters(locator_string, 0, len(locator_string))
		dh.endElement("fvalue")
		
	    dh.endElement("facet")

	dh.endElement("topicmap")
	
	dh.endDocument()


# ============================================================================
# Topic Map statistics
# ============================================================================

class TMStats:

    """Class that implements statistical and printing methods for TMs"""
    
    def __init__(self, writer=sys.stdout):

        """Initialize a new instance."""

	self.writer = writer

    def set_writer(self, writer):

	"""Set the output stream to use."""

	self.writer = writer
    
    # ----------------------------------------------------------------------------
    # Statistical and information printing methods
    # ----------------------------------------------------------------------------

    def print_tm_information(self, tm):

	"""Print complete information about the Topic Map."""

	self.print_tm_statistics(tm)
	
	self.print_tm_topics(tm)
	self.print_tm_associations(tm)
	self.print_tm_facets(tm)

    def print_tm_statistics(self, tm):

	"""Print statistics about the Topic Map."""

	self.print_header("STATISTICS", "=")
	self.writer.write("\n")
	self.writer.write(str(len(tm.get_topics())) +" topics\n")
	self.writer.write(str(len(tm.get_associations())) + " associations\n")
	self.writer.write(str(len(tm.get_facets())) + " facets\n")

    def print_tm_topics(self, tm):

	"""Print information about the topics in the Topic Map."""
	
	self.print_header("TOPICS", "=")
	if not tm.get_topics():
	    self.writer.write("No topics.\n")
	    return
	    
	# Topics
	for topic in tm.get_topics():
	    self.print_topic(topic, tm)

    def print_topic(self, topic, tm):

	"""Print information about this topic."""

	self.print_header("Topic: %s" % (topic.get_sgml_id()))
	self.writer.write("\n")

	# Topic types
	if topic.get_identity(): self.writer.write("  identity: %s\n" % (topic.get_identity()))
	if topic.get_types(): self.writer.write("  type(s): %s\n" % (topic.get_types()))
	self.writer.write("  type list: %s\n" % (tm.get_types_list(topic)))
	self.writer.write("  type hierarchy: %s\n" % (tm.get_types_hierarchy(topic, {})))
	
	# Topic names
	for name in topic.get_names():
	    if name.get_scopes(): self.writer.write("  scopes: %s\n" % (name.get_scopes()))
	    for basename in name.get_basenames():
		self.writer.write("  base name: %s\n" % (basename.get_name()))
		if basename.get_scopes():
		    self.writer.write("   scopes: %s\n" % (basename.get_scopes()))
	    for dispname in name.get_dispnames():
		self.writer.write("  display name: %s\n " % (dispname.get_name()))
		if dispname.get_scopes():
		    self.writer.write("   scopes: %s\n" % (dispname.get_scopes()))
	    for sortname in name.get_sortnames():
		self.writer.write("  sort name: %s\n" % (sortname.get_name()))
		if sortname.get_scopes():
		    self.writer.write("   scopes: %s\n" % (sortname.get_scopes()))

	# Topic occurences
	for occur in topic.get_occurrences():
	    self.writer.write("    Occurrence: %s\n" % (occur.get_rolename()))
	    self.writer.write("      scopes: %s\n" % (occur.get_scopes()))
	    for locator in occur.get_locators():
		self.writer.write("      Locator: %s\n" % (locator.get_loctype()))
		self.writer.write("        address: %s\n" % (locator.get_address()))

    def print_tm_associations(self, tm):

	"""Print information about the associations in the Topic Map."""

	self.print_header("ASSOCIATIONS", "=")
	if not tm.get_associations():
	    self.writer.write("No associations.\n")
	    return
	    
	# Associations
	for assoc in tm.get_associations():
	    self.print_association(assoc, tm)
	
    def print_association(self, association, tm):

	"""Print information about this association."""

	self.print_header("Association")
	self.writer.write("\n")

	if association.get_type(): self.writer.write("  types: %s\n" % (association.get_type()))
	if association.get_scopes(): self.writer.write("  scopes: %s\n" % (association.get_scopes()))
	for role in association.get_roles():
	    self.writer.write("  AssociationRole: %s\n" % (role.get_rolename()))
	    self.writer.write("    types: %s\n" % (role.get_type()))
	    for locator in role.get_locators():
		self.writer.write("    Located topic: %s\n" % (locator))

    def print_tm_facets(self, tm):

	"""Print information about the facets in the Topic Map."""

	self.print_header("FACETS", "=")
	if not tm.get_facets():
	    self.writer.write("No facets.\n")
	    return
	    
	# Facets
	for facet in tm.get_facets():
	    self.print_facet(facet, tm)

    def print_facet(self, facet, tm):

	"""Print information about this facet."""

	self.print_header("Facet")
	self.writer.write("\n")
	
	self.writer.write("  Type: %s\n" % (facet.get_type()))
	for fvalue in facet.get_values():
	    self.writer.write("  FacetValue: %s\n" % (fvalue.get_value()))
	    self.writer.write("    type: %s\n" % (fvalue.get_type()))
	    for locator in fvalue.get_locators():
		self.writer.write("      Located topic: %s\n" % (locator))

    def print_header(self, title, char="-", length=75):

	"""Print a header."""
	
	self.writer.write("%s\n" % (char * length))
	self.writer.write("%s\n" % (title))
